/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

const DBus       = imports.dbus;
const Gio        = imports.gi.Gio;
const Lang       = imports.lang;
const Main       = imports.ui.main;
const Mainloop   = imports.mainloop;
const Panel      = imports.ui.panel;
const PanelMenu  = imports.ui.panelMenu;
const PopupMenu  = imports.ui.popupMenu;
const Shell      = imports.gi.Shell;
const St         = imports.gi.St;
const Tweener    = imports.ui.tweener;

const PACKAGEKIT_BUS = "org.freedesktop.PackageKit";
const PACKAGEKIT_PATH = "/org/freedesktop/PackageKit";

const PackageKitIFace = {
    name: 'org.freedesktop.PackageKit',
    
    methods: [
        { name: 'CanAuthorize', inSignature: 's', outSignature: 's' },
        { name: 'GetTid', inSignature: '', outSignature: 's' },
        { name: 'GetTimeSinceAction', inSignature: 's', outSignature: 'u' },
        { name: 'GetTransactionList', inSignature: '', outSignature: 'as' },
        { name: 'StateHasChanged ', inSignature: 's', outSignature: '' },
        { name: 'SuggestDaemonQuit', inSignature: '', outSignature: '' },
        { name: 'GetDaemonState', inSignature: '', outSignature: 's' },
        { name: 'SetProxy', inSignature: 'ssssss', outSignature: '' },
        { name: 'SetRoot', inSignature: 's', outSignature: '' }
     ],
                
    signals: [
        { name: 'TransactionListChanged', inSignature: 'as' },
        { name: 'RestartSchedule', inSignature: '' },
        { name: 'RepoListChanged', inSignature: '' },
        { name: 'UpdatesChanged', inSignature: ''},
        { name: 'Changed', inSignature: ''}
     ],
             
    properties: [
        { name : 'VersionMajor', signature: 'u', access: 'read' },
        { name : 'VersionMinor', signature: 'u', access: 'read' },
        { name : 'VersionMicro', signature: 'u', access: 'read' },
        { name : 'BackendName', signature: 's',  access: 'read' },
        { name : 'BackendDescription', signature: 's', access: 'read' },
        { name : 'BackendAuthor', signature: 's', access: 'read' },
        { name : 'Roles', signature: 's', access: 'read' },
        { name : 'Groups', signature: 's', access: 'read' },
        { name : 'Filters', signature: 's', access: 'read' },
        { name : 'MimeTypes', signature: 's', access: 'read' },
        { name : 'Locked', signature: 'b', access: 'read' },
        { name : 'NetworkState', signature: 's', access: 'read' },
        { name : 'DistroId', signature: 's', access: 'read' }
    ]
};

const PackageKitTransactionIFace = {
    name: 'org.freedesktop.PackageKit.Transaction',
    
    methods: [
        { name: 'SetHints', inSignature: 'as', outSignature: '' },
        { name: 'AcceptEula', inSignature: 's', outSignature: '' },
        { name: 'Cancel', inSignature: '', outSignature: '' },
        { name: 'DownloadPackages', inSignature: 'bas', outSignature: '' },
        { name: 'GetCategories', inSignature: '', outSignature: '' },
        { name: 'GetDepends', inSignature: 'sasb', outSignature: '' },
        { name: 'GetDetails', inSignature: 'as', outSignature: '' },
        { name: 'GetFiles', inSignature: 'as', outSignature: '' },
        { name: 'GetOldTransactions', inSignature: 'u', outSignature: '' },
        { name: 'GetPackages', inSignature: 's', outSignature: '' },
        { name: 'GetRepoList', inSignature: 's', outSignature: '' },
        { name: 'GetRequires', inSignature: 'sasb', outSignature: '' },
        { name: 'GetUpdateDetail', inSignature: 'as', outSignature: '' },
        { name: 'GetUpdates', inSignature: 's', outSignature: '' },
        { name: 'GetDistroUpgrades', inSignature: '', outSignature: '' },
        { name: 'InstallFiles', inSignature: 'bas', outSignature: '' },
        { name: 'InstallPackages', inSignature: 'bas', outSignature: '' },
        { name: 'InstallSignature', inSignature: 'sss', outSignature: '' },
        { name: 'RefreshCache', inSignature: 'b', outSignature: '' },
        { name: 'RemovePackages', inSignature: 'asbb', outSignature: '' },
        { name: 'RepoEnable', inSignature: 'sb', outSignature: '' },
        { name: 'RepoSetData', inSignature: 'sss', outSignature: '' },
        { name: 'Resolve', inSignature: 'sas', outSignature: '' },
        { name: 'Rollback', inSignature: 's', outSignature: '' },
        { name: 'SearchDetails', inSignature: 'sas', outSignature: '' },
        { name: 'SearchFiles', inSignature: 'sas', outSignature: '' },
        { name: 'SearchGroups', inSignature: 'sas', outSignature: '' },
        { name: 'SearchNames', inSignature: 'sas', outSignature: '' },
        { name: 'SimulateInstallFiles', inSignature: 'as', outSignature: '' },
        { name: 'SimulateInstallPackages', inSignature: 'as', outSignature: '' },
        { name: 'SimulateRemovePackages', inSignature: 'asb', outSignature: '' },
        { name: 'SimulateUpdatePackages', inSignature: 'as', outSignature: '' },
        { name: 'UpdatePackages', inSignature: 'bas', outSignature: '' },
        { name: 'UpdateSystem', inSignature: 'b', outSignature: '' },
        { name: 'WhatProvides', inSignature: 'ssas', outSignature: '' },
        { name: 'UpgradeSystem', inSignature: 'ss', outSignature: '' }
     ],
                
    signals: [
        { name: 'Category', inSignature: 'sssss' },
        { name: 'Details', inSignature: 'ssssst' },
        { name: 'ErrorCode', inSignature: 'ss' },
        { name: 'Files', inSignature: 'ss' },
        { name: 'Finished', inSignature: 'su' },
        { name: 'Message', inSignature: 'ss' },
        { name: 'Package', inSignature: 'sss' },
        { name: 'RepoDetail', inSignature: 'ssb' },
        { name: 'RepoSignatureRequired', inSignature: 'ssssssss' },
        { name: 'EulaRequired', inSignature: 'ssss' },
        { name: 'MediaChangeRequired', inSignature: 'sss' },
        { name: 'RequireRestart', inSignature: 'ss' },
        { name: 'Transaction', inSignature: 'ssbsusus' },
        { name: 'UpdateDetail', inSignature: 'ssssssssssss' },
        { name: 'DistroUpgrade', inSignature: 'sss' },
        { name: 'Changed', inSignature: '' },
        { name: 'Destroy', inSignature: '' }
     ],
             
    properties: [
        { name : 'Role', signature: 's', access: 'read' },
        { name : 'Status', signature: 's', access: 'read' },
        { name : 'LastPackage', signature: 's', access: 'read' },
        { name : 'Uid', signature: 'u', access: 'read' },
        { name : 'Percentage', signature: 'u', access: 'read' },
        { name : 'Subpercentage', signature: 'u', access: 'read' },
        { name : 'AllowCancel', signature: 'b', access: 'read' },
        { name : 'CallerActive', signature: 'b', access: 'read' },
        { name : 'ElapsedTime', signature: 'u', access: 'read' },
        { name : 'RemainingTime', signature: 'u', access: 'read' },
        { name : 'Speed', signature: 'u', access: 'read' }
    ]
};

let PackageKit = DBus.makeProxyClass(PackageKitIFace);
let PackageKitTransaction = DBus.makeProxyClass(PackageKitTransactionIFace);

function MyPopupMenuItem() {
    this._init.apply(this, arguments);
}

MyPopupMenuItem.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function (text, params) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, params);

        this.label = new St.Label({ text: text });
        this.progressLabel = new St.Label();
        this.progressLabel.hide();
        this.spinner = new Panel.AnimatedIcon('process-working.svg', Panel.PANEL_ICON_SIZE);
		this.spinner.actor.hide();
		
		this.addActor(this.label);
		this.addActor(this.progressLabel);
		this.addActor(this.spinner.actor);
    },
    
    enableProgress: function() {
        this.progress = 0;
        this.progressLabel.set_text("0%");
        this.spinner.actor.show();
    },
    
    disableProgress: function() {
        this.progressLabel.hide();
        Tweener.addTween(this.spinner.actor, {
			opacity: 0,
			time: Panel.SPINNER_ANIMATION_TIME,
			transition: "easeOutQuad",
			onCompleteScope: this,
			onComplete: function() {
				this.spinner.actor.opacity = 255;
				this.spinner.actor.hide();
			}
		});
    },
    
    setProgress: function(p) {
        this.progressLabel.show();
        if(Number(p) >= this.progress) {
            this.progressLabel.set_text(p+"%");
            this.progress = Number(p);
        }
    }
};


function UpdateButton(settingsFile) {
	this._init(settingsFile);
}

UpdateButton.prototype = {
	__proto__: PanelMenu.Button.prototype,

	_init: function() {
		PanelMenu.Button.prototype._init.call(this, 0.0);
		
		this._settings = new Gio.Settings({ schema: 'org.gnome.shell.extensions.updateindicator' });
		this._settings.connect('changed::update-interval', Lang.bind(this, function() {
            this._stopTimeout();
            this._startTimeout();
		}));

        this._timeoutSource = -1;		
		this._pkgDaemon = new PackageKit(DBus.system, PACKAGEKIT_BUS, PACKAGEKIT_PATH);
	
		this._label = new St.Label({
			text : "0 Updates"
		});
		this.actor.add_actor(this._label);

		this._updatesMenuItem = new PopupMenu.PopupSubMenuMenuItem("0 Updates");

		this._refreshMenuItem = new MyPopupMenuItem("Refresh");
		this._refreshMenuItem.connect('activate', Lang.bind(this, this._actionRefresh));
		
		this._updateMenuItem = new MyPopupMenuItem("Update");
		this._updateMenuItem.connect('activate', Lang.bind(this, this._actionUpdate));
		
		this._upgradeMenuItem = new MyPopupMenuItem("Upgrade");
		this._upgradeMenuItem.connect('activate', Lang.bind(this, this._actionUpgrade));
		
		this._settingsMenuItem = new PopupMenu.PopupSubMenuMenuItem("Settings");
		
        let	item1 = new PopupMenu.PopupMenuItem("Update Interval:", {activate:false, reactive:false});
  		this._settingsMenuItem.menu.addMenuItem(item1);
        this._timeoutEntry = new St.Entry({ 
            text : this._settings.get_int('update-interval').toString(), 
            name : "TimeoutEntry"
        });
		item1.addActor(this._timeoutEntry);
		item1.addActor(new St.Label({ text : "seconds"}));
        let applyButton = new St.Button({
            style_class: 'notification-button',
            label     : 'Apply',
            reactive  : true,
            can_focus : true,
            });
        applyButton.connect('clicked', Lang.bind(this, function() {
            this._settings.set_int('update-interval', parseInt(this._timeoutEntry.get_text()));
        }));
        item1.addActor(applyButton);
		
		this.menu.addMenuItem(this._updatesMenuItem);
		this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
		this.menu.addMenuItem(this._refreshMenuItem);
    	this.menu.addMenuItem(this._updateMenuItem);
		this.menu.addMenuItem(this._upgradeMenuItem);
		this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
		this.menu.addMenuItem(this._settingsMenuItem);
		
		Main.panel._centerBox.add(this.actor);
		Main.panel._menus.addMenu(this.menu);
		
		this._actionRefresh();
		this._startTimeout();
	},
	
	_startTimeout: function() {
		this._timeoutSource = Mainloop.timeout_add_seconds(this._settings.get_int('update-interval'), Lang.bind(this, this._actionUpdate));
	},
	
	_stopTimeout: function() {
		if(this._timeoutSource >= 0) {
			Mainloop.source_remove(this._timeoutSource);
		}		
	},
	
	_startTransaction: function(callback, arguments) {

        this._pkgDaemon.GetTidRemote(DBus.CALL_FLAG_START, Lang.bind(this, function(tid, error) {
            if(error) {
                this._handlePackageKitError(error);
                return;
            }
            this._pkgTransaction = new PackageKitTransaction(DBus.system, PACKAGEKIT_BUS, tid);
            this._pkgTransaction.connect('Changed', Lang.bind(this, this._onChanged));
            this._pkgTransaction.connect('ErrorCode', Lang.bind(this, this._onErrorCode));
            callback.apply(this, arguments);
        }));
	},
    
	_actionRefresh : function() {
		this._setUILocked(true);
		this._updatesMenuItem.menu.removeAll();
		this._refreshMenuItem.enableProgress();
        this._updateMenuItem.disableProgress();
		this._upgradeMenuItem.disableProgress();
        this._updateItems = {};

        this._startTransaction(Lang.bind(this, function() {
            this._pkgTransaction.connect('Package', Lang.bind(this, this._onPackageUpdate));
            this._pkgTransaction.connect('Finished', Lang.bind(this, function() {
                let ids = new Array();
                for(key in this._updateItems) { ids.push( this._updateItems[key]['package_id'] ); }
                if(ids.length > 0) {
                    this._startTransaction(Lang.bind(this, function() {
                    
                        this._pkgTransaction.connect('UpdateDetail', Lang.bind(this, this._onUpdateDetail));
                        this._pkgTransaction.connect('Finished', Lang.bind(this, function() {
                            this._startTransaction(Lang.bind(this, function() {
                            
                                this._pkgTransaction.connect('Details', Lang.bind(this, this._onDetails));
                                this._pkgTransaction.connect('Finished', Lang.bind(this, function() {
                                    this._createUpdateMenuItems();
                                    this._refreshMenuItem.disableProgress();
                                    this._setUILocked(false);
                                }));
                                this._pkgTransaction.GetDetailsRemote(ids, Lang.bind(this, this._handlePackageKitError));
                            }));
                        }));
                        this._pkgTransaction.GetUpdateDetailRemote(ids, Lang.bind(this, this._handlePackageKitError));
                    }));
                } else {
                    this._createUpdateMenuItems();
                    this._refreshMenuItem.disableProgress();
                    this._setUILocked(false);
                }
            }));
            this._pkgTransaction.GetUpdatesRemote("none", Lang.bind(this, this._handlePackageKitError));
        }));
        return true;
	},
	
	_actionUpdate : function() {
    	this._setUILocked(true);
    	this._updateMenuItem.enableProgress();
        
    	this._startTransaction(Lang.bind(this, function() {
    	    this._pkgTransaction.connect('Finished', Lang.bind(this, function() {
    	        this._actionRefresh();
      	    }));
      	    this._pkgTransaction.RefreshCacheRemote(false, Lang.bind(this, this._handlePackageKitError));
    	}));
	},
	
	_actionUpgrade : function() {
    	this._setUILocked(true);
    	this._upgradeMenuItem.enableProgress();
    	
    	this._startTransaction(Lang.bind(this, function() {
    	    this._pkgTransaction.connect('Finished', Lang.bind(this, function() {
    	        this._actionRefresh();
      	    }));
      	    this._pkgTransaction.UpdateSystemRemote(false, Lang.bind(this, this._handlePackageKitError));
    	}));
	},
	
	_setUILocked : function(locked) {
		if(!locked) {
    		this._refreshMenuItem.actor.remove_style_pseudo_class('disabled');
			this._updatesMenuItem.actor.remove_style_pseudo_class('disabled');
			this._updateMenuItem.actor.remove_style_pseudo_class('disabled');
			this._upgradeMenuItem.actor.remove_style_pseudo_class('disabled');
			this._settingsMenuItem.actor.remove_style_pseudo_class('disabled');
		} else {
            this._refreshMenuItem.actor.add_style_pseudo_class('disabled');
			this._updatesMenuItem.actor.add_style_pseudo_class('disabled');
			this._updateMenuItem.actor.add_style_pseudo_class('disabled');
			this._upgradeMenuItem.actor.add_style_pseudo_class('disabled');
			this._settingsMenuItem.actor.add_style_pseudo_class('disabled');
		}
		this._refreshMenuItem.actor.reactive = !locked;
		this._updatesMenuItem.actor.reactive = !locked;
		this._updateMenuItem.actor.reactive = !locked;
		this._upgradeMenuItem.actor.reactive = !locked;
		this._settingsMenuItem.actor.reactive = !locked;
	},
	
	_onPackageUpdate: function(dummy, info, package_id, summary) {
	    let id = package_id.split(';', 1)[0]
    	this._updateItems[id] = {};
    	this._updateItems[id]['info'] = info;   
    	this._updateItems[id]['summary'] = summary;
    	this._updateItems[id]['package_id'] = package_id;
	},
	
	_onDetails: function(dummy, package_id, license, group, detail, url, size) {
	    let id = package_id.split(';', 1)[0];
        this._updateItems[id]['group'] = group;
	    this._updateItems[id]['detail'] = detail;
	    this._updateItems[id]['size'] = size;
	},
	
	_onUpdateDetail: function(dummy, package_id, update, obsolete, vendor_url, bugzilla_url, cve_url, restart, update_text, changelog, state, issued, updated) {
	    let id = package_id.split(';', 1)[0];
	    this._updateItems[id]['update_text'] = update_text;
	    this._updateItems[id]['changelog'] = changelog;
	},
	
	_onChanged: function() {
        this._pkgTransaction.GetAllRemote(Lang.bind(this, function() {
            if(arguments[1] != null) {
                Main.notifyError("Error", arguments[1].message)
            } else if(Number(arguments[0]['Percentage']) != 101){
                if(arguments[0]['Status'] == "refresh-cache") {
                    this._updateMenuItem.setProgress(arguments[0]['Percentage']);
                } else if(arguments[0]['Status'] == "update-system") {
                    this._upgradeMenuItem.setProgress(arguments[0]['Percentage']);
                }
            }
        }));
	},
	
	_onErrorCode: function(dummy, code, details) {
	    Main.notifyError("UpdateIndicator:", details);
	},
	
	_handlePackageKitError: function(error) {
    	if(error) {
        	Main.notifyError("UpdateIndicator", error.message);
        }
	},
	
	_createUpdateMenuItems: function() {
	    let cnt = 0;
	    for(key in this._updateItems) {
	        cnt++;
	        let val = this._updateItems[key];
	        let item = new PopupMenu.PopupMenuItem(key, { activate:false });
		    if("blocked" == val['info']) {
		        item.actor.add_style_pseudo_class('blocked');
		    }
		    
  		    let tooltip = "%{summary}\n%{name} (Size: %{size} MiB)\nPriority: %{priority}\n\nChanges:\n%{changelog}\n\n%{text}";
  		    tooltip = tooltip.replace(/%{summary}/, val['summary']);
  		    tooltip = tooltip.replace(/%{name}/, key);
  		    tooltip = tooltip.replace(/%{priority}/, val['info']);
  		    tooltip = tooltip.replace(/%{changelog}/, val['changelog']);
  		    tooltip = tooltip.replace(/%{text}/, val['update_text']);
		    
		    let size = parseInt(val['size']);
		    if(size != 0) {
		        size /= (1024*1024);
		    }
		    tooltip = tooltip.replace(/%{size}/, String(size));
   		    
   		    item.actor.set_tooltip_text(tooltip);
		    this._updatesMenuItem.menu.addMenuItem(item);
	    }
	    
	    let label = cnt + " Updates";
	    this._label.set_text(label);
	    this._updatesMenuItem.label.set_text(label);
	},
	
	destroy: function() {
		// TODO: kill button
	}
};

// old API
function main(extensionMeta) {
	_button = new UpdateButton();
	//Main.panel.addToStatusArea('drive-menu', _button);
}

// new API
let updateButton;

function init(extensionMeta) {
}

function disable() {
	updateButton.destroy();
}

function enable() {
	updateButton = new UpdateButton();
}
